##########################################################################
# RSArmageddon - RSA cryptography and cryptoanalysis toolkit             #
# Copyright (C) 2020,2021                                                #
# Vittorio Mignini a.k.a. M1gnus <vittorio.mignini@gmail.com>            #
# Simone Cimarelli a.k.a. Aquilairreale <aquilairreale@ymail.com>        #
#                                                                        #
# This program is free software: you can redistribute it and/or modify   #
# it under the terms of the GNU General Public License as published by   #
# the Free Software Foundation, either version 3 of the License, or      #
# (at your option) any later version.                                    #
#                                                                        #
# This program is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
# GNU General Public License for more details.                           #
#                                                                        #
# You should have received a copy of the GNU General Public License      #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
##########################################################################

import sys

from pathlib import Path
from functools import partial

from ..args import args
from ..utils import output
from ..crypto import cipher, uncipher
from ..utils import int_from_path, output_text, compute_d, compute_n, compute_pubkey, complete_privkey, DEFAULT_E


def run():
    n, e, d, p, q, phi = args.n, args.e, args.d, args.p, args.q, args.phi
    if args.command == "encrypt":
        n, e = compute_pubkey(n, e, d, p, q, phi)
        f = partial(cipher, n=n, e=e)
    else:
        if e is None:
            e = DEFAULT_E
        d = compute_d(n, e, d, p, q, phi)
        n = compute_n(n, e, d, p, q, phi)
        f = partial(uncipher, n=n, e=e, d=d)

    if not args.inputs:
        output.error("Nothing to do")

    for text, base_filename in args.inputs: # TODO: implement args.inputs parsing func
        if isinstance(text, Path):
            text = int_from_path(text)
        elif isinstance(text, bytes):
            text = int.from_bytes(text, "big")
        first_line = True
        for std in args.encryption_standard:
            filename = base_filename
            if filename is not True and len(std) > 1:
                filename = f"{filename}.{std}"
            try:
                output_data = f(text, std=std)
            except ValueError as e:
                output.error(f"{std}: {e}")
                continue
            encoding = None
            if args.command == "decrypt":
                encoding = args.encoding
                label = "plaintext"
            else:
                label = "ciphertext"
            if len(args.encryption_standard) > 1:
                if not first_line:
                    output.newline()
                first_line = False
                output.info(f"Using encryption standard {std}")
            output_text(label, output_data, filename, encoding=encoding, json_output=args.json) # TODO: add enc std label for stdout



##########################################################################
# RSArmageddon - RSA cryptography and cryptoanalysis toolkit             #
# Copyright (C) 2020,2021                                                #
# Vittorio Mignini a.k.a. M1gnus <vittorio.mignini@gmail.com>            #
# Simone Cimarelli a.k.a. Aquilairreale <aquilairreale@ymail.com>        #
#                                                                        #
# This program is free software: you can redistribute it and/or modify   #
# it under the terms of the GNU General Public License as published by   #
# the Free Software Foundation, either version 3 of the License, or      #
# (at your option) any later version.                                    #
#                                                                        #
# This program is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
# GNU General Public License for more details.                           #
#                                                                        #
# You should have received a copy of the GNU General Public License      #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
##########################################################################

import sys
import binascii

from pathlib import Path
from functools import partial
from gmpy2 import invert
from Crypto.Cipher import PKCS1_OAEP, PKCS1_v1_5
from Crypto.PublicKey import RSA

from .utils import to_bytes_auto, DEFAULT_E


standards = {
    "oaep": PKCS1_OAEP,
    "pkcs": PKCS1_v1_5
}


def cipher(m, n, e=None, std="pkcs"):
    """Encrypt a plaintext using RSA

    Arguments:
    m -- plaintext
    n -- RSA modulus
    e -- RSA public exponent
    std -- encryption standard
    """

    if e is None:
        e = DEFAULT_E

    if m > n-2:
        e_mess = str(m)
        if len(e_mess) > 10:
            e_mess = f"{e_mess[:10]} ..."
        raise ValueError(f"Modulus to small for plaintext: {e_mess}")

    if std == "raw":
        return pow(m, e, n)

    key = RSA.construct((n, e))
    standard = standards[std]

    encryptor = standard.new(key)

    return int.from_bytes(encryptor.encrypt(to_bytes_auto(m)), "big")


def uncipher(c, n, e=None, d=None, std="pkcs"):
    """Decrypt a plaintext using RSA

    Arguments:
    c -- ciphertext
    n -- RSA modulus
    d -- RSA private exponent
    std -- encryption standard
    """

    if e is None:
        e = DEFAULT_E

    c %= n

    if std == "raw":
        return pow(c, d, n)

    key = RSA.construct((n, e, d))
    standard = standards[std]

    decryptor = standard.new(key)

    if standard is PKCS1_v1_5:
        decryptor.decrypt = partial(decryptor.decrypt, sentinel=None)

    dec = decryptor.decrypt(to_bytes_auto(c))
    if dec is None:
        raise ValueError(f"Invalid ciphertext (encryption standard: {std})")
    return int.from_bytes(dec, "big")



##########################################################################
# RSArmageddon - RSA cryptography and cryptoanalysis toolkit             #
# Copyright (C) 2020,2021                                                #
# Vittorio Mignini a.k.a. M1gnus <vittorio.mignini@gmail.com>            #
# Simone Cimarelli a.k.a. Aquilairreale <aquilairreale@ymail.com>        #
#                                                                        #
# This program is free software: you can redistribute it and/or modify   #
# it under the terms of the GNU General Public License as published by   #
# the Free Software Foundation, either version 3 of the License, or      #
# (at your option) any later version.                                    #
#                                                                        #
# This program is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
# GNU General Public License for more details.                           #
#                                                                        #
# You should have received a copy of the GNU General Public License      #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
##########################################################################

import os
import sys
import re
import shutil
import subprocess

from pathlib import Path, PurePosixPath
from itertools import count, chain
from subprocess import Popen, PIPE, TimeoutExpired
from psutil import Process, wait_procs

from .utils import output


if os.name == "nt":
    from winreg import HKEY_CURRENT_USER, OpenKey, EnumKey, QueryValueEx


INSTALL_SAGE_POSIX = """We could not seem to find a functioning SageMath installation on your system.
Please refer to these online instructions to fix this: <TODO: insert link to appropriate github readme paragraph>"""
INSTALL_SAGE_NT = """We could not seem to find a functioning SageMath installation on your system.
Please refer to these online instructions to fix this: <TODO: insert link to appropriate github readme paragraph>"""
NO_JAVA = "This program will not run on Jython and other Java based execution environments"

SUPPORTED_VMAJ = 9


def best_version(versions):
    supported = [(vmaj, vmin) for vmaj, vmin in versions if vmaj == SUPPORTED_VMAJ]
    if supported:
        return max(supported)
    return max(versions)


def cyg_path(path, cyg_runtime):
    path = Path(path)
    if cyg_runtime is None:
        return PurePosixPath(path)
    cyg_runtime = Path(cyg_runtime)
    p = subprocess.run(
            [str(cyg_runtime/"bin"/"cygpath.exe"), str(path)],
            stdout=PIPE, text=True) # TODO: is Unicode handled properly?
    return PurePosixPath(p.stdout[:-1])


def cyg_bash(cyg_runtime):
    if cyg_runtime is None:
        return []
    cyg_runtime = Path(cyg_runtime)
    return [str(cyg_runtime/"bin"/"bash.exe"), "--norc", "--login"]


version_re = re.compile(r"(\d+)\.(\d+)", re.ASCII)

def version(sage, cyg_runtime=None):
    sage = PurePosixPath(sage)
    p = subprocess.run(
            [*cyg_bash(cyg_runtime), str(sage), "--version"],
            stdout=PIPE, text=True)
    m = version_re.search(p.stdout)
    if not m:
        raise RuntimeError(f"SageMath version could not be identified (not a sage executable?). Version string: {p.stdout}")
    vmaj, vmin = m.group(1, 2)
    return int(vmaj), int(vmin)


def get_sage_nt_locations_registry():
    assert os.name == "nt"
    sage_key_re = re.compile(r"SageMath-(\d+)\.(\d+)", re.IGNORECASE)
    with OpenKey(HKEY_CURRENT_USER, r"SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall") as root_key:
        try:
            for i in count():
                subkey_name = EnumKey(root_key, i)
                m = sage_key_re.search(subkey_name)
                if m:
                    vmaj, vmin = m.group(1, 2)
                    with OpenKey(root_key, subkey_name) as subkey:
                        location, _ = QueryValueEx(subkey, "InstallLocation")
                    yield Path(location).resolve(), vmaj, vmin
        except OSError:
            pass


def get_sage_nt_locations_appdata():
    assert os.name == "nt"
    try:
        appdata = Path(os.environ.get("LOCALAPPDATA", os.environ["APPDATA"]))
    except KeyError as e:
        raise RuntimeError("Cannot find the AppData folder! Is this a sane Windows machine?") from e
    for subdir in appdata.glob("sagemath*"):
        m = version_re.search(subdir.name)
        if m:
            vmaj, vmin = m.group(1, 2)
            yield subdir.resolve(), vmaj, vmin


def get_sage_nt():
    assert os.name == "nt"
    sages_by_ver = {}
    for location, vmaj, vmin in chain(
            get_sage_nt_locations_registry(), get_sage_nt_locations_appdata()):
        cyg_runtime = location/"runtime"
        sage = PurePosixPath(f"/opt/sagemath-{vmaj}.{vmin}/local/bin/sage")
        try:
            ver = version(sage, cyg_runtime)
        except (RuntimeError, OSError):
            continue
        sages_by_ver.setdefault(ver, set()).add((sage, cyg_runtime))

    if not sages_by_ver:
        raise RuntimeError(INSTALL_SAGE_NT)

    best_ver = best_version(sages_by_ver.keys())
    return (best_ver, *sages_by_ver[best_ver].pop())


def get_sage_posix():
    sage = shutil.which("sage")
    if sage is None:
        raise RuntimeError(INSTALL_SAGE_POSIX)
    return version(sage), sage, None


def get_sage_java():
    raise RuntimeError(NO_JAVA)


sage = None
cyg_runtime = None

def get_sage():
    global sage, cyg_runtime
    if sage is not None:
        return sage, cyg_runtime
    get_sage_by_platform = {
        "posix": get_sage_posix,
        "nt": get_sage_nt,
        "java": get_sage_java
    }
    version, sage, cyg_runtime = get_sage_by_platform[os.name]()
    vmaj, vmin = version
    if vmaj != SUPPORTED_VMAJ:
        output.warning(f"Using unsupported SageMath version {vmaj}.{vmin}")
        output.warning(f"RSArmageddon is not supposed to work with versions other than {SUPPORTED_VMAJ}.x,")
        output.warning(f"try installing the latest one of those before reporting a bug")
    return sage, cyg_runtime


def run(script, *args, env=None, timeout=None):
    sage, cyg_runtime = get_sage()
    p = Popen(
            [*cyg_bash(cyg_runtime), str(sage), str(cyg_path(script, cyg_runtime)), *args],
            stdout=PIPE, env=env, text=True)
    try:
        output, _ = p.communicate(timeout=timeout)
    except TimeoutExpired as e:
        pp = Process(p.pid)
        subprocesses = [pp, *pp.children(recursive=True)]
        for subp in subprocesses:
            subp.terminate()
        wait_procs(subprocesses)
        raise e
    return p, output



##########################################################################
# RSArmageddon - RSA cryptography and cryptoanalysis toolkit             #
# Copyright (C) 2020,2021                                                #
# Vittorio Mignini a.k.a. M1gnus <vittorio.mignini@gmail.com>            #
# Simone Cimarelli a.k.a. Aquilairreale <aquilairreale@ymail.com>        #
#                                                                        #
# This program is free software: you can redistribute it and/or modify   #
# it under the terms of the GNU General Public License as published by   #
# the Free Software Foundation, either version 3 of the License, or      #
# (at your option) any later version.                                    #
#                                                                        #
# This program is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
# GNU General Public License for more details.                           #
#                                                                        #
# You should have received a copy of the GNU General Public License      #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
##########################################################################

import sys

from encodings.aliases import aliases
from contextlib import redirect_stdout

from . import attacks
from .utils import output


banner_text = r"""
  ______  _____  ___                                      _     _
  | ___ \/  ___|/ _ \                                    | |   | |
  | |_/ /\ `--./ /_\ \_ __ _ __ ___   __ _  __ _  ___  __| | __| | ___  _ __
  |    /  `--. \  _  | '__| '_ ` _ \ / _` |/ _` |/ _ \/ _` |/ _` |/ _ \| '_ \
  | |\ \ /\__/ / | | | |  | | | | | | (_| | (_| |  __/ (_| | (_| | (_) | | | |
  \_| \_|\____/\_| |_/_|  |_| |_| |_|\__,_|\__, |\___|\__,_|\__,_|\___/|_| |_|
                                            __/ |
                                           |___/

  Written by M1gnus && AquilaIrreale -- PGIATASTI
"""

credits_text = """\
  Glory to PGiatasti:
      - Vittorio aka M1gnus
      - Alessio aka Alexius
      - Cristiano aka ReverseBrain
      - Riccardo aka ODGrip
      - Nalin aka Lotus
      - Emanuele aka KaiserSource
      - Federico aka Heichou
      - Antonio aka CoffeeStraw
      - Giacomo aka Giaxo
      - Andrei aka S0uND_0f_s1lence
      - Simone aka AquilaIrreale

  https://pgiatasti.it/ -- Visit our site to find out more about us
"""


def print_banner():
    output.yellow(banner_text)


def print_credits():
    output.white(credits_text)


def print_attacks_short():
    all_attacks = set(attacks.builtin)
    all_attacks.update(attacks.installed)
    for attack in sorted(all_attacks):
        print(attack)


def _print_attacks(header, attacks):
    output.white(f" {header} ".center(32, "="))
    output.white(f"==={str(len(attacks)).center(26)}===")
    output.white("="*32)
    print()

    for attack in sorted(attacks):
        output.white(attack)


def print_attacks():
    with redirect_stdout(sys.stderr):
        _print_attacks("Builtin", attacks.builtin)
        if attacks.installed:
            print()
            _print_attacks("Installed", attacks.installed)
        print()


def print_encodings():
    for enc in sorted(c for c in set(aliases.values()) if not c.endswith("_codec")):
        print(enc)


def version():
    print("RSArmageddon v2.0 --Ares--")



##########################################################################
# RSArmageddon - RSA cryptography and cryptoanalysis toolkit             #
# Copyright (C) 2020,2021                                                #
# Vittorio Mignini a.k.a. M1gnus <vittorio.mignini@gmail.com>            #
# Simone Cimarelli a.k.a. Aquilairreale <aquilairreale@ymail.com>        #
#                                                                        #
# This program is free software: you can redistribute it and/or modify   #
# it under the terms of the GNU General Public License as published by   #
# the Free Software Foundation, either version 3 of the License, or      #
# (at your option) any later version.                                    #
#                                                                        #
# This program is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
# GNU General Public License for more details.                           #
#                                                                        #
# You should have received a copy of the GNU General Public License      #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
##########################################################################

import os
import sys

from textwrap import dedent
from functools import partial
from pathlib import Path
from argparse import ArgumentParser, Action, Namespace, SUPPRESS, RawDescriptionHelpFormatter

from .certs import load_key
from .parsing import (
        parse_int_arg,
        parse_time,
        parse_list,
        parse_std_list,
        path_or_stdout)


help_formatter = partial(RawDescriptionHelpFormatter, max_help_position=30)

main_description = dedent("""\
        RSArmageddon: Smashing RSA for fun and profit

        Commands:
          attack                      execute attacks
          pem                         manage keys and key files
          encrypt                     encrypt data and files
          decrypt                     decrypt data and files
          factor                      factorize a number with PARI factorization
          ecm                         factorize a number with ellyptic curve method
          isprime                     primality test
          eulerphi                    calculate euler phi of a number""")

attack_description = dedent("""\
        Attack weak public keys and recover private keys""")

pem_description = dedent("""\
        Manage keys and key files""")

cipher_description = dedent("""\
        Encrypt data and files using different encryption standards""")

uncipher_description = dedent("""\
        Decrypt data and files using different encryption standards""")

factor_description = dedent("""\
        Factorize a number using PARI factorization""")

ecm_description = dedent("""\
        Factorize a number using elliptic curve method""")

isprime_description = dedent("""\
        Primality test""")

eulerphi_description = dedent("""\
        Calculate Euler's phi function of a number""")

epilog = dedent("""\
        Number format:
            All numbers can be input in a variety of formats and bases.
            RSArmageddon understands regular base 10 numbers and python literals
            introduced by 0x for hex, 0o for octal or 0b for binary.
            Other less common bases can be specified in the form number:base
            where base is either an integer between 2 and 32, b64 for base64, or
            b85 for base85""")


class ReadKeyFile(Action):
    def __call__(self, parser, namespace, path, option_string=None):
        key = load_key(path)
        for k, v in zip(("n", "e", "d", "p", "q"), key):
            if getattr(namespace, k, None) is None:
                setattr(namespace, k, v)


key_parser = ArgumentParser(add_help=False)
key_parser.add_argument("--key", "-k", action=ReadKeyFile, type=Path,          metavar="FILE",   help="Path to a key file")
key_parser.add_argument("-n",          action="store",     type=parse_int_arg, metavar="NUMBER", help="RSA public modulus")
key_parser.add_argument("-e",          action="store",     type=parse_int_arg, metavar="NUMBER", help="RSA public exponent")
key_parser.add_argument("-d",          action="store",     type=parse_int_arg, metavar="NUMBER", help="RSA private exponent")
key_parser.add_argument("-p",          action="store",     type=parse_int_arg, metavar="NUMBER", help="RSA first prime factor")
key_parser.add_argument("-q",          action="store",     type=parse_int_arg, metavar="NUMBER", help="RSA second prime factor")
key_parser.add_argument("--phi",       action="store",     type=parse_int_arg, metavar="NUMBER", help="Euler's phi of RSA public modulus")


class Input(Action):
    def __call__(self, parser, namespace, input_, option_string=None):
        inputs = getattr(namespace, "inputs")
        inputs.append((input_, True))


class Output(Action):
    def __call__(self, parser, namespace, dest, option_string=None):
        inputs = getattr(namespace, "inputs", None)
        if not inputs:
            raise ValueError(f"{option_string} found but no inputs have been given yet") from e
        cur_text, _ = inputs[-1]
        inputs[-1] = (cur_text, dest)


text_parser_common = ArgumentParser(add_help=False)
text_parser_common.add_argument("--encryption-standard", "--std", action="store", type=parse_std_list, default=["pkcs"], metavar="STD", help="Comma-separated list. Try to use these encryption standards when reading or writing ciphertext files (default: pkcs)")
text_parser_common.set_defaults(inputs=[])

plaintext_parser = ArgumentParser(add_help=False, parents=[text_parser_common])
plaintext_parser.add_argument("--plaintext",      "--pt",  "--encrypt",      action=Input,  type=parse_int_arg,  metavar="NUMBER",        help="Use number as plaintext")
plaintext_parser.add_argument("--plaintext-raw",  "--ptr", "--encrypt-raw",  action=Input,  type=os.fsencode,    metavar="BINARY_STRING", help="Use string as plaintext")
plaintext_parser.add_argument("--plaintext-file", "--ptf", "--encrypt-file", action=Input,  type=Path,           metavar="FILE",          help="Use file as plaintext")
plaintext_parser.add_argument("--output",         "-o",                      action=Output, type=path_or_stdout, metavar="FILE",          help="Output ciphertext to file or stdout (applies to previous input)")

ciphertext_parser = ArgumentParser(add_help=False, parents=[text_parser_common])
ciphertext_parser.add_argument("--encoding", action="store", help="")
ciphertext_parser.add_argument("--ciphertext",      "--ct",  "--decrypt",      action=Input,  type=parse_int_arg,  metavar="NUMBER",        help="Use number as ciphertext")
ciphertext_parser.add_argument("--ciphertext-raw",  "--ctr", "--decrypt-raw",  action=Input,  type=os.fsencode,    metavar="BINARY_STRING", help="Use string as ciphertext")
ciphertext_parser.add_argument("--ciphertext-file", "--ctf", "--decrypt-file", action=Input,  type=Path,           metavar="FILE",          help="Use file as ciphertext")
ciphertext_parser.add_argument("--output",          "-o",                      action=Output, type=path_or_stdout, metavar="FILE",          help="Output plaintext to file or stdout (applies to previous input)")

commons_parser = ArgumentParser(argument_default=SUPPRESS, add_help=False)
commons_parser.add_argument("--show-attacks",            action="store_const", const=True, help="Show all available attacks")
commons_parser.add_argument("--show-attacks-short",      action="store_const", const=True, help="Show all available attacks (short form)")
commons_parser.add_argument("--show-encodings",          action="store_const", const=True, help="Show all available encodings")
commons_parser.add_argument("--credits",                 action="store_const", const=True, help="Show credits")
commons_parser.add_argument("--version",                 action="store_const", const=True, help="Show version")
commons_parser.add_argument("--json",                    action="store_const", const=True, help="Turn on json output format")
commons_parser.add_argument("--quiet", "--silent", "-s", action="store_const", const=True, help="Suppress informative output")
commons_parser.add_argument("--color",                choices=["auto", "always", "never"], help="Set color output behavior")

main_parser = ArgumentParser(parents=[commons_parser], formatter_class=help_formatter, description=main_description)

scripts_parser = ArgumentParser(add_help=False)
scripts_parser.add_argument("n", action="store", type=parse_int_arg, metavar="NUMBER", help="Input number")

command_subparsers = main_parser.add_subparsers(dest="command")
attack_parser   = command_subparsers.add_parser("attack",  parents=[commons_parser, ciphertext_parser],             formatter_class=help_formatter, description=attack_description,   epilog=epilog)
pem_parser      = command_subparsers.add_parser("pem",     parents=[commons_parser, key_parser],                    formatter_class=help_formatter, description=pem_description,      epilog=epilog)
cipher_parser   = command_subparsers.add_parser("encrypt", parents=[commons_parser, key_parser, plaintext_parser],  formatter_class=help_formatter, description=cipher_description,   epilog=epilog)
uncipher_parser = command_subparsers.add_parser("decrypt", parents=[commons_parser, key_parser, ciphertext_parser], formatter_class=help_formatter, description=uncipher_description, epilog=epilog)
command_subparsers.add_parser("factor",                    parents=[commons_parser, scripts_parser],                formatter_class=help_formatter, description=factor_description,   epilog=epilog)
command_subparsers.add_parser("ecm",                       parents=[commons_parser, scripts_parser],                formatter_class=help_formatter, description=ecm_description,      epilog=epilog)
command_subparsers.add_parser("isprime",                   parents=[commons_parser, scripts_parser],                formatter_class=help_formatter, description=isprime_description,  epilog=epilog)
command_subparsers.add_parser("eulerphi",                  parents=[commons_parser, scripts_parser],                formatter_class=help_formatter, description=eulerphi_description, epilog=epilog)

pem_parser.add_argument("--generate", "-g",                      action="store_true",                                 help="Generate a new 2048 bit key pair")
pem_parser.add_argument("--dump-values", "--dumpvalues", "--dv", action="store_true",                                 help="Dump key values to standard output")
pem_parser.add_argument("--create-public", "--cpu",              action="store", type=path_or_stdout, metavar="FILE", help="Output public key to file")
pem_parser.add_argument("--create-private", "--cpr",             action="store", type=path_or_stdout, metavar="FILE", help="Output private key to file")
pem_parser.add_argument("--file-format", "--ff",                 choices=["pem", "der", "openssh"], default="pem",    help="Set output key file format")


class NewKey(Action):
    def __call__(self, parser, namespace, n, option_string=None):
        keys = getattr(namespace, "keys")
        keys.append((n, None))


class SetE(Action):
    def __call__(self, parser, namespace, e, option_string=None):
        keys = getattr(namespace, "keys", None)
        if not keys:
            raise ValueError(f"-e found but no moduli (-n) have been given yet")
        n, _ = keys[-1]
        keys[-1] = (n, e)


attack_parser.add_argument("attacks",                    action="store",                                type=parse_list,     metavar="ATTACKS_LIST", help="Comma-separated list of attacks to be executed")
attack_parser.add_argument("-n",                         action=NewKey,                                 type=parse_int_arg,  metavar="NUMBER",       help="Target key RSA public modulus")
attack_parser.add_argument("-e",                         action=SetE,                                   type=parse_int_arg,  metavar="NUMBER",       help="Target key RSA public exponent (applies to preceding -n)")
attack_parser.add_argument("--n-e-file", "--nef",        action="append", dest="n_e_files", default=[], type=Path,           metavar="FILE",         help="Read target public keys from text FILE, one comma-separated n,e pair per line")
attack_parser.add_argument("--key", "-k",                action="append", dest="key_paths", default=[], type=Path,           metavar="PATH",         help="Read target public keys from PATH. PATH can either be a single key file or a directory containing key files")
attack_parser.add_argument("--exts", "-x",               action="store",  default=["pem", "pub"],       type=parse_list,     metavar="EXTENSIONS",   help="Comma-separated list of file extensions. Selects which files are picked up by -k when PATH is a directory")
attack_parser.add_argument("--recursive", "-r",          action="store_true",                                                                        help="Descend recursively inside directories when looking for key files (-k only looks into the first level by default)")
attack_parser.add_argument("--output-key", "--ok",       action="store_true",                                                                        help="Output first cracked key to standard output")
attack_parser.add_argument("--output-key-file", "--okf", action="store",                                type=path_or_stdout, metavar="FILE",         help="Output first cracked key to FILE")
attack_parser.add_argument("--output-key-dir", "--okd",  action="store",                                type=Path,           metavar="DIRECTORY",    help="Output all cracked keys to this directory")
attack_parser.add_argument("--timeout", "-t",            action="store",                                type=parse_time,     metavar="TIME",         help="Set maximum run time allowed for each attack")
attack_parser.set_defaults(keys=[])


class CustomNamespace(Namespace):
    def __init__(self):
        super().__init__()
        self.show_attacks=False
        self.show_attacks_short=False
        self.show_encodings=False
        self.credits=False
        self.version=False
        self.json=False
        self.quiet=False
        self.color="auto"


args = main_parser.parse_args(namespace=CustomNamespace())

if __name__ == "__main__":
    print(args)



##########################################################################
# RSArmageddon - RSA cryptography and cryptoanalysis toolkit             #
# Copyright (C) 2020,2021                                                #
# Vittorio Mignini a.k.a. M1gnus <vittorio.mignini@gmail.com>            #
# Simone Cimarelli a.k.a. Aquilairreale <aquilairreale@ymail.com>        #
#                                                                        #
# This program is free software: you can redistribute it and/or modify   #
# it under the terms of the GNU General Public License as published by   #
# the Free Software Foundation, either version 3 of the License, or      #
# (at your option) any later version.                                    #
#                                                                        #
# This program is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
# GNU General Public License for more details.                           #
#                                                                        #
# You should have received a copy of the GNU General Public License      #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
##########################################################################

import os
import re
from pathlib import Path
from importlib import resources
from importlib.resources import contents, is_resource

from . import builtin_attacks


if os.name == "posix":
    try:
        cfg_dir = Path(os.environ["XDG_CONFIG_HOME"])
    except KeyError:
        cfg_dir = Path.home()/".config"
    user_atk_dir = cfg_dir/"rsarmageddon"/"attacks"
    user_atk_dir.mkdir(parents=True, exist_ok=True)
    sys_atk_dir = Path("/usr/share/rsarmageddon/attacks")
    path = [user_atk_dir, sys_atk_dir]
elif os.name == "nt":
    app_data = Path(os.environ.get("LOCALAPPDATA", os.environ["APPDATA"]))
    user_atk_dir = app_data/"RSArmageddon"/"attacks"
    user_atk_dir.mkdir(parents=True, exist_ok=True)
    path = [user_atk_dir]
else:
    path = []


suffix_re = re.compile(r"\.sage$")
prefix_re = re.compile(r"^\d{2}_")


def attack_name(s):
    if isinstance(s, Path):
        s = s.stem
    else:
        s = suffix_re.sub("", s)
    return prefix_re.sub("", s).replace("\n", " ")


builtin = {
    attack_name(res): res
    for res in sorted(contents(builtin_attacks))
    if is_resource(builtin_attacks, res)
        and res.endswith(".sage")
}


installed = {
    attack_name(f): f
    for d in reversed(path)
    for f in sorted(d.glob("*.sage"))
    if f.is_file()
}


def attack_path(name):
    try:
        return installed[name]
    except KeyError:
        pass
    try:
        return resources.path(builtin_attacks, builtin[name])
    except KeyError:
        pass
    raise ValueError(f"There is no attack named '{name}'")





#!/usr/bin/env python3

##########################################################################
# RSArmageddon - RSA cryptography and cryptoanalysis toolkit             #
# Copyright (C) 2020,2021                                                #
# Vittorio Mignini a.k.a. M1gnus <vittorio.mignini@gmail.com>            #
# Simone Cimarelli a.k.a. Aquilairreale <aquilairreale@ymail.com>        #
#                                                                        #
# This program is free software: you can redistribute it and/or modify   #
# it under the terms of the GNU General Public License as published by   #
# the Free Software Foundation, either version 3 of the License, or      #
# (at your option) any later version.                                    #
#                                                                        #
# This program is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
# GNU General Public License for more details.                           #
#                                                                        #
# You should have received a copy of the GNU General Public License      #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
##########################################################################

import os
import sys

from itertools import compress

from .utils import output
from .args import args
from .banner import (
        print_banner,
        print_credits,
        print_attacks,
        print_attacks_short,
        print_encodings,
        version)
from .commands import pem, ciphertool, attack, misc


def main():
    output.init(args.color)

    if args.quiet:
        sys.stderr.close()
        sys.stderr = open(os.devnull, "w")

    if not any((args.show_attacks_short, args.show_encodings, args.version)):
        print_banner()

    banner_actions = compress(*zip(
        (version, args.version),
        (print_credits, args.credits),
        (print_attacks, args.show_attacks),
        (print_attacks_short, args.show_attacks_short),
        (print_encodings, args.show_encodings)
    ))

    try:
        action = next(banner_actions)
    except StopIteration:
        pass
    else:
        action()
        return

    if not args.command:
        return

    commands = {
        "pem": pem.run,
        "encrypt": ciphertool.run,
        "decrypt": ciphertool.run,
        "attack": attack.run
    }

    command = commands.get(args.command, misc.run)

    try:
        command()
    except (ValueError, OSError) as e:
        output.error(e)


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        output.error("Interrupted")
    #except Exception as e:
    #    output.error(f"Unhandled exception: {e}")
    else:
        sys.exit(0)

    sys.exit(1)



##########################################################################
# RSArmageddon - RSA cryptography and cryptoanalysis toolkit             #
# Copyright (C) 2020,2021                                                #
# Vittorio Mignini a.k.a. M1gnus <vittorio.mignini@gmail.com>            #
# Simone Cimarelli a.k.a. Aquilairreale <aquilairreale@ymail.com>        #
#                                                                        #
# This program is free software: you can redistribute it and/or modify   #
# it under the terms of the GNU General Public License as published by   #
# the Free Software Foundation, either version 3 of the License, or      #
# (at your option) any later version.                                    #
#                                                                        #
# This program is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
# GNU General Public License for more details.                           #
#                                                                        #
# You should have received a copy of the GNU General Public License      #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
##########################################################################

import sys
import signal
import multiprocessing

from operator import itemgetter
from functools import wraps
from itertools import count, islice
from contextlib import redirect_stdout

from . import output


name = None
_default_key_name = None


def init(attack_name, default_key_name, *, min_keys=1, min_ciphertexts=0, deduplicate=None):
    global name, _default_key_name

    name = attack_name
    _default_key_name = default_key_name

    def excepthook(exctype, value, traceback):
        if exctype in (KeyboardInterrupt, RuntimeError):
            sys.exit(2)
        else:
            sys.__excepthook__(exctype, value, traceback)
    sys.excepthook = excepthook

    ciphertexts = []
    keys = []
    color = "auto"

    with open(sys.argv[1], "r", encoding="ascii") as f:
        for line in f:
            line = line.strip()
            if not line or line.isspace():
                continue
            kind, _, line = line.partition(":")
            if kind == "k":
                n, e, keyname = line.split(",", maxsplit=2)
                keys.append((int(n), int(e), keyname or None))
            elif kind == "c":
                text, textname = line.split(",", maxsplit=1)
                ciphertexts.append((int(text), textname))
            elif kind == "C":
                color = line.strip()
            else:
                raise ValueError("Unexpected input type '{}' from input file".input(kind))

    output.init(color)

    if deduplicate:
        if deduplicate in ("n", "ns"):
            cmp_key = itemgetter(0)
        elif deduplicate == "keys":
            cmp_key = itemgetter(0, 1)
        else:
            raise ValueError("Bad value for deduplicate argument '{}'".format(deduplicate))
        keys = {cmp_key(key): key for key in reversed(keys)}
        keys = list(keys.values())
        keys.reverse()
        if len(keys) < min_keys:
            fail("This attack needs at least {} distinct keys".format(min_keys))
    else:
        if len(keys) < min_keys:
            fail("This attack needs at least {} keys".format(min_keys))

    if len(ciphertexts) < min_ciphertexts:
        fail("This attack needs at least {} cihertexts".format(min_cihertexts))

    output.success("{} attack started".format(name))
    return ciphertexts, keys


@wraps(multiprocessing.Pool)
def Pool(*args, **kwargs):
    old_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
    pool = multiprocessing.Pool(*args, **kwargs)
    signal.signal(signal.SIGINT, old_handler)
    return pool


def with_name_set(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        if name is None:
            raise RuntimeError("Attacks must call set_name before using this function")
        return f(*args, **kwargs)
    return wrapper


_keys = []
@with_name_set
def keys(*keys):
    _keys.extend(keys)
    for _, _, d, p, q, *_ in keys:
        for name, value in (("d", d), ("p", p), ("q", q)):
            if value is not None:
                info("{}: {}".format(name, value))


_cleartexts = []
@with_name_set
def cleartexts(*cleartexts):
    _cleartexts.extend(cleartexts)


@with_name_set
def success():
    unnamed_keys = sum(1 for key in _keys if len(key) == 5 or key[-1] is None)
    field_width = len(str(unnamed_keys))
    auto_name = (f"{_default_key_name}_{c:0{field_width}}" for c in count())
    for key in _keys:
        if not isinstance(key, (tuple, list)) or len(key) not in (5, 6):
            raise ValueError("Bad key '{}'".format(key))
        if len(key) == 5 or key[-1] is None:
            key = (*key[:5], next(auto_name))
        print("k:{}".format(",".join(str(x) if x is not None else "" for x in key)))
    for cleartext in _cleartexts:
        if isinstance(cleartext, int):
            text, textname = str(cleartext), ""
        elif isinstance(cleartext, tuple):
            text, textname = cleartext
            text = str(text)
            textname = textname if textname is not None else ""
        else:
            raise ValueError("Bad cleartext '{}'".format(cleartext))
        print("c:{},{}".format(text, textname))
    output.success("{} attack succeeded".format(name))
    sys.exit(0)


@with_name_set
def fail(*s, bad_key=False):
    if s:
        output.error(" ".join(map(str, s)))
    output.error("{} attack failed".format(name))
    sys.exit(1 if not bad_key else 2)


@with_name_set
def info(*s):
    if s:
        output.primary(" ".join(map(str, s)))
    else:
        output.newline()


def positive_int(s):
    i = int(s)
    if i <= 0:
        raise ValueError("Must be a positive number")
    return i


_input = input
@with_name_set
def input(prompt=None, *, default=None, validator=None):
    if prompt is not None:
        prompt_default = " [{}]".format(default) if default is not None else ""
        prompt = "{}{}: ".format(prompt, prompt_default)

    if validator is None:
        validator = lambda x: x

    while True:
        if prompt is not None:
            output.info(prompt, newline=False)

        try:
            inp = _input().strip()
        except EOFError:
            if prompt:
                output.newline()
            fail()

        if not inp:
            if default is not None:
                return default
            else:
                output.warning("Must enter a value")
                continue

        try:
            inp = validator(inp)
        except ValueError as e:
            output.warning("Invalid input ({})".format(e))
        else:
            return inp



##########################################################################
# RSArmageddon - RSA cryptography and cryptoanalysis toolkit             #
# Copyright (C) 2020,2021                                                #
# Vittorio Mignini a.k.a. M1gnus <vittorio.mignini@gmail.com>            #
# Simone Cimarelli a.k.a. Aquilairreale <aquilairreale@ymail.com>        #
#                                                                        #
# This program is free software: you can redistribute it and/or modify   #
# it under the terms of the GNU General Public License as published by   #
# the Free Software Foundation, either version 3 of the License, or      #
# (at your option) any later version.                                    #
#                                                                        #
# This program is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
# GNU General Public License for more details.                           #
#                                                                        #
# You should have received a copy of the GNU General Public License      #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
##########################################################################

import os
import sys
import colorama

from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from itertools import chain
from contextlib import redirect_stdout
from subprocess import TimeoutExpired

from .. import sage
from .. import utils
from .. import attack_lib
from ..args import args
from ..certs import encode_privkey, load_key, load_keys
from ..crypto import uncipher
from ..attacks import attack_path, builtin, installed
from ..parsing import parse_n_e_file
from ..utils import (
        output, DEFAULT_E, to_bytes_auto, output_text,
        compute_d, complete_privkey, int_from_path,
        compute_d, copy_resource, copy_resource_tree)


def parse_output(s):
    cleartexts = []
    keys = []
    for line in s.splitlines():
        line = line.strip()
        if not line:
            continue
        kind, _, value = map(str.strip, line.partition(":"))
        if kind == "c":
            text, _, file = value.partition(",")
            cleartexts.append((int(text), Path(file) if file else True))
        elif kind == "k":
            *key, name = value.split(",")
            key = tuple(int(x) if x else None for x in key)
            keys.append((key, name or None))
        else:
            raise ValueError(f"Unexpected return type '{kind}' from sage script")
    return cleartexts, keys


def run():
    attacks = list(dict.fromkeys(args.attacks))
    if len(attacks) != len(args.attacks):
        output.warning("Attacks specified more than once are ignored")
    try:
        i = attacks.index("all")
    except ValueError:
        pass
    else:
        l = attacks[:i]
        r = attacks[i+1:]
        all_attacks = {
            **installed,
            **builtin
        }
        for attack in chain(l, r):
            all_attacks.pop(attack, None)
        attacks = [*l, *all_attacks, *r]

    keys = []
    for n, e in args.keys:
        if e is None:
            e = DEFAULT_E
        keys.append(((n, e), None))
    for path in args.n_e_files:
        keys.extend((key, None) for key in parse_n_e_file(path))
    for path in args.key_paths:
        if path.is_dir():
            keys.extend(load_keys(path, recursive=args.recursive, exts=args.exts))
        else:
            n, e, _, _, _ = load_key(path)
            keys.append(((n, e), path.name))

    if not keys:
        output.error("please provide at least one key")
        return

    with TemporaryDirectory() as attack_lib_dir, \
            NamedTemporaryFile("w", encoding="ascii") as input_file:
        with redirect_stdout(input_file):
            print(f"C:{args.color}")
            for (n, e), name in keys:
                print(f"k:{n},{e},{name if name is not None else ''}")
            for text, name in args.inputs:
                if isinstance(text, Path):
                    text = int_from_path(text)
                elif isinstance(text, bytes):
                    text = int.from_bytes(text)
                print(f"c:{text},{name if name is not True else ''}")
        input_file.flush()

        copy_resource(attack_lib, "attack.py", attack_lib_dir)
        copy_resource(utils, "output.py", attack_lib_dir)
        copy_resource_tree(colorama, attack_lib_dir)

        _, cyg_runtime = sage.get_sage()

        env = os.environ.copy()
        env["PYTHONPATH"] = str(sage.cyg_path(attack_lib_dir, cyg_runtime))

        for attack in attacks:
            try:
                script_manager = attack_path(attack)
            except ValueError as e:
                output.error(e)
                continue

            with script_manager as script:
                try:
                    p, script_output = sage.run(script, input_file.name, env=env, timeout=args.timeout)
                except TimeoutExpired:
                    output.warning(f"Timeout expired for attack {attack}")
                    continue

                if p.returncode == 2: # attack determined the key is bad (i.e. not an RSA key)
                    break

                if p.returncode: # attack failed for other reasons
                    continue

                cleartexts, keys = parse_output(script_output)

                if cleartexts:
                    output.info("Plaintext recovered")
                    for text, file in cleartexts:
                        output_text("plaintext", text, file, encoding=args.encoding, json_output=args.json)
                        if file is True:
                            print()

                if len(keys) == 1:
                    key, _ = keys[0]
                    n, e, d, p, q = key
                    if d is None:
                        d = compute_d(*key)
                        key = (n, e, d, p, q)

                    if args.output_key_file is None and args.output_key:
                        args.output_key_file = True

                    if args.output_key_file is True:
                        sys.stdout.buffer.write(encode_privkey(*key, "PEM"))
                        print()
                    elif args.output_key_file:
                        with open(args.output_key_file, "wb") as f:
                            f.write(encode_privkey(*key, "PEM"))

                    for text, filename in args.inputs:
                        if isinstance(text, Path):
                            with open(text, "rb") as f:
                                text_bytes = f.read()
                                text = int.from_bytes(text_bytes, "big")
                        else:
                            text_bytes = to_bytes_auto(text)
                        for std in args.encryption_standard:
                            output.newline()
                            output.info(f"Decrypting 0x{text_bytes.hex()} with encryption standard {std}")
                            try:
                                cleartext = uncipher(text, n, e, d, std)
                            except ValueError as e:
                                output.error(e)
                            else:
                                output_text("plaintext", cleartext, filename, encoding=args.encoding, json_output=args.json)
                            if filename is True:
                                print()

                if args.output_key_dir is not None:
                    for key, name in keys:
                        key = complete_privkey(*key)
                        with open(args.output_key_dir/f"{name}.pem", "wb") as f:
                            f.write(encode_privkey(*key, "PEM"))

                if keys:
                    break



##########################################################################
# RSArmageddon - RSA cryptography and cryptoanalysis toolkit             #
# Copyright (C) 2020,2021                                                #
# Vittorio Mignini a.k.a. M1gnus <vittorio.mignini@gmail.com>            #
# Simone Cimarelli a.k.a. Aquilairreale <aquilairreale@ymail.com>        #
#                                                                        #
# This program is free software: you can redistribute it and/or modify   #
# it under the terms of the GNU General Public License as published by   #
# the Free Software Foundation, either version 3 of the License, or      #
# (at your option) any later version.                                    #
#                                                                        #
# This program is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
# GNU General Public License for more details.                           #
#                                                                        #
# You should have received a copy of the GNU General Public License      #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
##########################################################################

import sys
import json

from contextlib import redirect_stdout
from pathlib import Path

from gmpy2 import invert
from Crypto.PublicKey import RSA

from .utils import output, DEFAULT_E


common_formats = {
    ".pem": "PEM",
    ".der": "DER",
    ".openssh": "OpenSSH"
}


private_formats = {}
private_formats.update(common_formats)

public_formats = {".pub": "PEM"}
public_formats.update(common_formats)


def infer_format_priv(p):
    """Infer private key file format from extension

    Arguments:
    p -- path to key file
    """
    try:
        return private_formats[p.suffix] if isinstance(p, Path) else None
    except KeyError:
        return None


def infer_format_pub(p):
    """Infer public key file format from extension

    Arguments:
    p -- path to key file
    """
    try:
        return public_formats[p.suffix] if isinstance(p, Path) else None
    except KeyError:
        return None


def generate_key(e=None):
    """Generate a new key randomly

    Keyword Arguments:
    e -- RSA public exponent
    """
    e = e if e is not None else DEFAULT_E
    key = RSA.generate(2048, e=e)
    return key.n, key.e, key.d, key.p, key.q


def print_key(n=None, e=None, d=None, p=None, q=None, dp=None, dq=None, pinv=None, qinv=None):
    """Print key elements in a more readable format

    Keyword arguments:
    n -- RSA public modulus
    e -- RSA public exponent
    d -- RSA private exponent
    p -- first factor
    q -- second factor
    dp
    dq
    pinv
    qinv
    file -- file-like object
    """
    output.primary(f"n: {n}")
    output.primary(f"e: {e}")
    output.primary(f"d: {d}")
    output.primary(f"p: {p}")
    output.primary(f"q: {q}")
    output.newline()
    output.secondary(f"dp: {dp}")
    output.secondary(f"dq: {dq}")
    output.secondary(f"pinv: {pinv}")
    output.secondary(f"qinv: {qinv}")
    output.newline()


def print_key_json(n=None, e=None, d=None, p=None, q=None, dp=None, dq=None, pinv=None, qinv=None):
    d = {k: str(v) for k, v in locals().items()}
    json.dump(d, sys.stdout, indent=4)
    print()


def load_key(path):
    """Load key elements from a key file

    Arguments:
    path -- path to a key file
    """

    # obtain a RsaKey object from the key file: https://pycryptodome.readthedocs.io/en/latest/src/public_key/rsa.html 
    keyfile = open(path, "rb")
    key = RSA.importKey(keyfile.read())

    # If the key is a private key dump the private key values
    if key.has_private():
        return key.n, key.e, key.d, key.p, key.q
    else:
        return key.n, key.e, None, None, None


def load_keys(path, exts=("pem", "pub"), recursive=False):
    """Load key elements from keys found in a directory

    Arguments:
    path -- path to keys directory
    exts -- iterable of file extensions
    """
    keys = []
    path = Path(path)

    for ext in exts:
        if not ext.startswith("."):
            ext = f".{ext}"

        pattern = f"*{ext}"
        if recursive:
            pattern = f"**/{pattern}"

        for key_file in path.glob(pattern):
            n, e, _, _, _ = load_key(key_file)
            keys.append(((n, e), key_file.stem))

    return keys


file_formats = {
    "pem": "PEM",
    "der": "DER",
    "openssh": "OpenSSH"
}


def encode_pubkey(n, e, file_format):
    """Encode a public key to file_format

    Arguments:
    n -- RSA modulus
    e -- RSA public exponent
    file_format -- file format
    """
    file_format = file_formats[file_format.casefold()]
    try:
        return RSA.construct((n, e)).exportKey(format=file_format)
    except NotImplementedError as e:
        raise ValueError(f"Cannot create a public key file (key data is incomplete): {(n, e)}") from e


def encode_privkey(n, e, d, p, q, file_format):
    """Encode a public key to file_format

    Arguments:
    n -- RSA modulus
    e -- RSA public exponent
    d -- RSA private exponent
    p -- RSA first factor
    q -- RSA second factor
    file_format -- file format
    """
    file_format = file_formats[file_format.casefold()]
    try:
        return RSA.construct((n, e, d, p, q)).exportKey(format=file_format)
    except NotImplementedError as exc:
        raise ValueError(f"Cannot create a private key file (key data is incomplete): {(n, e, d, p, q)}") from exc



##########################################################################
# RSArmageddon - RSA cryptography and cryptoanalysis toolkit             #
# Copyright (C) 2020,2021                                                #
# Vittorio Mignini a.k.a. M1gnus <vittorio.mignini@gmail.com>            #
# Simone Cimarelli a.k.a. Aquilairreale <aquilairreale@ymail.com>        #
#                                                                        #
# This program is free software: you can redistribute it and/or modify   #
# it under the terms of the GNU General Public License as published by   #
# the Free Software Foundation, either version 3 of the License, or      #
# (at your option) any later version.                                    #
#                                                                        #
# This program is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
# GNU General Public License for more details.                           #
#                                                                        #
# You should have received a copy of the GNU General Public License      #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
##########################################################################

import sys

from ..args import args
from ..utils import output
from ..certs import print_key, print_key_json, generate_key, encode_pubkey, encode_privkey, load_key
from ..utils import compute_extra_key_elements, compute_pubkey, complete_privkey


def run():
    if args.generate:
        n, e, d, p, q = generate_key()
    else:
        n, e, d, p, q = args.n, args.e, args.d, args.p, args.q

        try:
            n, e = compute_pubkey(n, e, d, p, q)
            n, e, d, p, q = complete_privkey(n, e, d, p, q)
        except ValueError:
            pass

    if args.dump_values:
        dp, dq, pinv, qinv = compute_extra_key_elements(d, p, q)
        if args.json:
            print_key_json(n, e, d, p, q, dp, dq, pinv, qinv)
        else:
            print_key(n, e, d, p, q, dp, dq, pinv, qinv)

    if args.create_public:
        key = encode_pubkey(n, e, args.file_format)
        if args.create_public is True:
            sys.stdout.buffer.write(key)
            print()
        else:
            with open(args.create_public, "wb") as f:
                f.write(key)

    if args.create_private:
        key = encode_privkey(n, e, d, p, q, args.file_format)
        if args.create_private is True:
            sys.stdout.buffer.write(key)
            print()
        else:
            with open(args.create_private, "wb") as f:
                f.write(key)

    if not any((args.dump_values, args.create_public, args.create_private)):
        output.error("Nothing to do")



from pathlib import Path
from setuptools import setup

def sh_escape(f):
    f = str(f).replace("'", r"'\''")
    return f"'{f}'"

print("cleaning up compiled sage files")
for f in Path("rsarmageddon").glob("**/*.sage.py"):
    if not f.is_dir():
        f.unlink(missing_ok=True)
        print(f"removed {sh_escape(f)}")

setup()



##########################################################################
# RSArmageddon - RSA cryptography and cryptoanalysis toolkit             #
# Copyright (C) 2020,2021                                                #
# Vittorio Mignini a.k.a. M1gnus <vittorio.mignini@gmail.com>            #
# Simone Cimarelli a.k.a. Aquilairreale <aquilairreale@ymail.com>        #
#                                                                        #
# This program is free software: you can redistribute it and/or modify   #
# it under the terms of the GNU General Public License as published by   #
# the Free Software Foundation, either version 3 of the License, or      #
# (at your option) any later version.                                    #
#                                                                        #
# This program is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
# GNU General Public License for more details.                           #
#                                                                        #
# You should have received a copy of the GNU General Public License      #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
##########################################################################

import re
import sys
import binascii
import subprocess

from base64 import b64decode, b85decode
from pathlib import Path

from .crypto import standards
from .utils import DEFAULT_E


def parse_unsigned(s, base=None):
    """Convert to int raising ValueError on negative values

    Arguments:
    s -- string to convert

    Keyword arguments:
    base -- numeric base of the conversion
    """
    if base is not None:
        if base < 2:
            raise ValueError("base cannot be less than 2")
        ret = int(s, base)
    else:
        s = s.strip()
        if len(s) > 1 and not s[1].isalpha():
            s = s.lstrip("0")
        if not s:
            s = "0"
        ret = int(s, 0)
    if ret < 0:
        raise ValueError(f"'{ret}' is not unsigned")
    return ret


def parse_int_arg(s):
    """Take a string in the format number[:base] or any format
    recognized by python and return the corresponding integer

    Arguments:
    s -- string to convert
    """
    number, sep, base = s.rpartition(":")
    if not sep:
        number, base = base, number
    number = number.strip()
    base = base.strip()
    try:
        if base in ("b64", "base64"):
            return int.from_bytes(b64decode(number, validate=True), "big")
        elif base in ("b85", "base85"):
            return int.from_bytes(b85decode(number), "big")
    except binascii.Error as e:
        raise ValueError(str(e)) from e

    if base:
        return parse_unsigned(number, int(base))
    else:
        return parse_unsigned(number)


time_re1 = re.compile(r"(\d+)([hms]?)")
time_re2 = re.compile(r"(?:(?:(\d+):)?(?:(\d+):))?(\d+)")
time_mult = {
    "": 1,
    "s": 1,
    "m": 60,
    "h": 3600
}
def parse_time(s):
    sc = s.strip().casefold()
    m1 = time_re1.fullmatch(sc)
    if m1:
        return int(m1.group(1)) * time_mult[m1.group(2)]
    m2 = time_re2.fullmatch(sc)
    if m2:
        h, m, s = m2.groups()
        h = h or "0"
        m = m or "0"
        h, m, s = int(h), int(m), int(s)
        return h*3600 + m*60 + s
    raise ValueError(f"Bad time '{s}'")


def parse_list(s):
    """Parse a comma separated list of arguments (example1,example2,...) into a list

    Arguments:
    s -- comma separated string
    """
    return [x if x else None for x in s.split(",")] if s else []


def parse_int_list(s):
    """Parse a comma separated list of ints (example1,example2,...) into a list

    Arguments:
    s -- comma separated string
    """
    return [parse_int_arg(x) if x is not None else None for x in parse_list(s)]


def parse_std_list(s):
    all_standards = ["raw", *standards.keys()]
    allowed = {*all_standards, "all"}
    l = [x.strip().casefold() for x in s.split(",") if x.strip()]
    l = list(dict.fromkeys(l)) # Deduplicate list keeping order
    for std in l:
        if std not in allowed:
            raise ValueError(f"Invalid encryption standard '{std}'")
    if "all" in l:
        l = all_standards
    return l


def path_or_stdout(s):
    if s == "-":
        return True
    return Path(s)


def parse_n_e_file(filename):
    with open(filename, "r") as f:
        for line in f:
            line = line.strip()
            if not line:
                continue
            n, _, e = line.partition(",")
            yield (parse_int_arg(n), parse_int_arg(e) if e else DEFAULT_E)



##########################################################################
# RSArmageddon - RSA cryptography and cryptoanalysis toolkit             #
# Copyright (C) 2020,2021                                                #
# Vittorio Mignini a.k.a. M1gnus <vittorio.mignini@gmail.com>            #
# Simone Cimarelli a.k.a. Aquilairreale <aquilairreale@ymail.com>        #
#                                                                        #
# This program is free software: you can redistribute it and/or modify   #
# it under the terms of the GNU General Public License as published by   #
# the Free Software Foundation, either version 3 of the License, or      #
# (at your option) any later version.                                    #
#                                                                        #
# This program is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
# GNU General Public License for more details.                           #
#                                                                        #
# You should have received a copy of the GNU General Public License      #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
##########################################################################

import sys
import colorama
from colorama import Fore


def _print(*args, **kwargs):
    print(*args, **kwargs, file=sys.stderr, flush=True)


def init(color="auto"):
    if color == "auto":
        colorama.init()
    elif color == "always":
        colorama.init(strip=False)
    elif color == "never":
        colorama.init(strip=True, convert=False)
    else:
        raise ValueError(f"Bad color setting {color!r}")


def yellow(msg, newline=True):
    _print(f"{Fore.YELLOW}{msg}{Fore.RESET}", end=("\n" if newline else ""))


def white(msg, newline=True):
    _print(f"{Fore.LIGHTWHITE_EX}{msg}{Fore.RESET}", end=("\n" if newline else ""))


def success(msg, newline=True):
    _print(f"[{Fore.GREEN}+{Fore.RESET}] {msg}", end=("\n" if newline else ""))


def primary(msg, newline=True):
    _print(f"[{Fore.CYAN}*{Fore.RESET}] {msg}", end=("\n" if newline else ""))


def secondary(msg, newline=True):
    _print(f"[{Fore.MAGENTA}#{Fore.RESET}] {msg}", end=("\n" if newline else ""))


def info(msg, newline=True):
    _print(f"[{Fore.BLUE}${Fore.RESET}] {msg}", end=("\n" if newline else ""))


def warning(msg, newline=True):
    _print(f"[{Fore.LIGHTYELLOW_EX}W{Fore.RESET}] {msg}", end=("\n" if newline else ""))


def error(msg, newline=True):
    _print(f"[{Fore.RED}-{Fore.RESET}] {msg}", end=("\n" if newline else ""))


def newline():
    _print()


if __name__ == "__main__":
    yellow("Yellow text")
    white("White text")
    success("Success")
    primary("Primary output")
    secondary("Secondary output")
    info("Info")
    warning("Warning")
    error("Error")



##########################################################################
# RSArmageddon - RSA cryptography and cryptoanalysis toolkit             #
# Copyright (C) 2020,2021                                                #
# Vittorio Mignini a.k.a. M1gnus <vittorio.mignini@gmail.com>            #
# Simone Cimarelli a.k.a. Aquilairreale <aquilairreale@ymail.com>        #
#                                                                        #
# This program is free software: you can redistribute it and/or modify   #
# it under the terms of the GNU General Public License as published by   #
# the Free Software Foundation, either version 3 of the License, or      #
# (at your option) any later version.                                    #
#                                                                        #
# This program is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
# GNU General Public License for more details.                           #
#                                                                        #
# You should have received a copy of the GNU General Public License      #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
##########################################################################

import os
import sys
import subprocess
import colorama

from importlib import resources
from tempfile import TemporaryDirectory

from .. import sage
from .. import scripts
from .. import utils

from ..args import args
from ..utils import copy_resource, copy_resource_tree


def run():
    script_names = {
        "eulerphi": "euler_phi.sage",
        "factor":   "factor.sage",
        "ecm":      "ecm_factor.sage",
        "isprime":  "isprime.sage"
    }

    script_name = script_names[args.command]

    with resources.path(scripts, script_name) as script, \
            TemporaryDirectory() as libs_dir:
        copy_resource(utils, "output.py", libs_dir)
        copy_resource_tree(colorama, libs_dir)

        _, cyg_runtime = sage.get_sage()

        env = os.environ.copy()
        env["PYTHONPATH"] = str(sage.cyg_path(libs_dir, cyg_runtime))

        sage.run(script, str(args.n), args.color, env=env)



##########################################################################
# RSArmageddon - RSA cryptography and cryptoanalysis toolkit             #
# Copyright (C) 2020,2021                                                #
# Vittorio Mignini a.k.a. M1gnus <vittorio.mignini@gmail.com>            #
# Simone Cimarelli a.k.a. Aquilairreale <aquilairreale@ymail.com>        #
#                                                                        #
# This program is free software: you can redistribute it and/or modify   #
# it under the terms of the GNU General Public License as published by   #
# the Free Software Foundation, either version 3 of the License, or      #
# (at your option) any later version.                                    #
#                                                                        #
# This program is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of         #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
# GNU General Public License for more details.                           #
#                                                                        #
# You should have received a copy of the GNU General Public License      #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
##########################################################################

import sys
import json
import random
import hashlib

from pathlib import Path
from shutil import copyfileobj
from functools import partial
from contextlib import redirect_stdout
from importlib import resources, import_module
from base64 import b64encode, urlsafe_b64encode
from gmpy2 import invert, isqrt, gcd

from . import output


DEFAULT_E = 65537


def carmichael_lcm(p, q):
    phi = (p-1)*(q-1)
    return phi // gcd(p-1, q-1)


def byte_length(n):
    """Return byte length of the given integer

    Arguments:
    n -- int
    """
    return -(n.bit_length() // -8)


def to_bytes_auto(n):
    """Convert int to shortest possible bytes object, big endian

    Arguments:
    n -- int
    """
    return n.to_bytes(byte_length(n), "big")


def int_from_path(path):
    with open(path, "rb") as f:
        return int.from_bytes(f.read(), "big")


def compute_extra_key_elements(d, p, q):
    """Compute extra key elements

    Arguments:
    d -- RSA private exponent
    p -- RSA first factor
    q -- RSA second factor
    """
    if d is not None:
        dp = d%(p-1)
        dq = d%(q-1)
    else:
        dp, dq = None, None
    if p is not None and q is not None:
        pinv = int(invert(p, q))
        qinv = int(invert(q, p))
    else:
        pinv, qinv = None, None

    return dp, dq, pinv, qinv


def compute_pubkey(n, e, d, p, q, phi=None):
    """Compute public key elements

    Arguments:
    n -- RSA modulus
    e -- RSA public exponent
    d -- RSA private exponent
    p -- RSA first factor
    q -- RSA second factor
    """

    tup = (n, e, d, p, q)

    pks = set()

    if n is not None and e is not None:
        pks.add((n, e))

    if n is not None and d is not None:
        if phi is not None:
            tmp_e = int(invert(d, phi))
            pks.add((n, tmp_e))
        else:
            if p is None:
                p = q
            if p is not None:
                q = n//p
                phi =  (p-1) * (q-1)
                tmp_e = int(invert(d, phi))
                if e is not None and tmp_e != e:
                    tmp_e = int(invert(d, phi//gcd(p-1,q-1)))
                pks.add((n, tmp_e))

    if p is not None and q is not None:
        if d is not None:
            phi =  (p-1) * (q-1)
            tmp_e = int(invert(d, phi))
            if e is not None and tmp_e != e:
                tmp_e = int(invert(d, phi//gcd(p-1,q-1)))
            pks.add((p*q, tmp_e))
        if e is not None:
            pks.add((p*q, e))

    if len(pks) != 1:
        raise ValueError(f"Inconsistent parameters {tup}")

    return pks.pop()


def recover_pq(n, e, d):
    k = d*e - 1

    if k % 2 != 0:
        raise ValueError(f"p and q cannot be recovered from these parameters {(n, e, d)}")

    factor = 1
    while k % 2 == 0:
        factor *= 2
        k //= 2

    for i in range(100):
        b = 0
        g = random.randint(0, n)
        y = pow(g, k, n)
        if y == 1 or y == (n - 1):
            continue
        for j in range(1, factor):
            x = pow(y, 2, n)
            if x in (1, n-1):
                break
            y = x
        else:
            x = pow(y, 2, n)
        if x == 1:
            break
    else:
        raise ValueError(f"p and q cannot be recovered (not enough iterations?)")

    p = int(gcd(y-1, n))
    q = n // p
    return p, q


def complete_privkey(n, e, d, p, q, phi=None, use_lcm=True):
    """Compute missing private key elements

    Arguments:
    n -- RSA modulus
    e -- RSA public exponent
    d -- RSA private exponent
    p -- RSA first factor
    q -- RSA second factor
    """

    tup = (n, e, d, p, q, phi)

    if n is None and (p is None or q is None):
        raise ValueError(f"You have to provide n or both p and q in tuple '{tup}'")
    if n is not None and (p is None and q is None and phi is None) and (d is None or e is None):
        raise ValueError(f"If you provide n, you must provide also either one of p, q or phi, or d and e, in tuple '{tup}'")
    if e is None and d is None:
        raise ValueError(f"You have to provide e or d in tuple '{tup}'")

    if n is not None and p is None and q is None and phi is not None:
        p = ((n + 1 - phi) - isqrt((n + 1 - phi)**2 - 4*n)) // 2

    if n is None:
        n = p*q
    elif p is None:
        p = n//q
    elif q is None:
        q = n//p

    if p is None:
        p, q = recover_pq(n, e, d)

    if n != p * q:
        raise ValueError(f"n is not equal to p * q in tuple '{tup}'")

    if use_lcm:
        phi = carmichael_lcm(p, q)
    else:
        phi = (p-1) * (q-1)

    if e is None:
        e = int(invert(d, phi))
    elif d is None:
        d = int(invert(e, phi))

    return n, e, d, p, q

def compute_d(n, e, d, p, q, phi=None):
    """Compute d from available parameters

    Arguments:
    n -- RSA modulus
    e -- RSA public exponent
    d -- RSA private exponent
    p -- RSA first factor
    q -- RSA second factor
    """

    tup = (n, e, d, p, q, phi)

    ds = set()

    if d is not None:
        ds.add(d)

    if e is None and len(ds) == 0:
        raise ValueError("Missing public exponent")

    if phi is not None:
        ds.add(int(invert(e, phi)))

    if p is None:
        p = q

    if p is not None:
        if q is not None:
            tmp_d = int(invert(e, (p-1) * (q-1)))
            if d is not None and tmp_d != d:
                tmp_d = int(invert(e, carmichael_lcm(p, q)))
            ds.add(tmp_d)
        if n is not None:
            q = n//p
            tmp_d = int(invert(e, (p-1) * (q-1)))
            if d is not None and tmp_d != d:
                tmp_d = int(invert(e, carmichael_lcm(p, q)))
            ds.add(tmp_d)

    if len(ds) != 1:
        raise ValueError(f"Inconsistent parameters {tup}")

    return ds.pop()


def compute_n(n, e, d, p, q, phi=None):
    """Compute d from available parameters

    Arguments:
    n -- RSA modulus
    e -- RSA public exponent
    d -- RSA private exponent
    p -- RSA first factor
    q -- RSA second factor
    """

    tup = (n, e, d, p, q, phi)

    ns = set()

    if n is not None:
        ns.add(n)

    if p is not None and q is not None:
        ns.add(p*q)

    if p is None:
        p = q

    if phi is not None and p is not None:
        ns.add((pow(p, 2) - p + p*phi) // (p-1))

    if len(ns) != 1:
        raise ValueError(f"Inconsistent parameters {tup}")

    return ns.pop()


def output_text(label, text, filename, encoding=None, json_output=False):
    text_raw = to_bytes_auto(text)
    text_b64 = b64encode(text_raw).decode("ascii")
    text_b64_url = urlsafe_b64encode(text_raw).decode("ascii")
    if encoding is not None:
        try:
            text_str = text_raw.decode(encoding)
        except ValueError:
            text_str = f"Cannot decode ({encoding})"
    else:
        text_str = None
    if filename is True:
        text_hex = f"0x{text_raw.hex()}"
        if json_output:
            output_obj = {
                "dec": str(text),
                "hex": text_hex,
                "raw": str(text_raw),
                "b64": str(text_b64),
                "url": str(text_b64_url)
            }
            if text_str is not None:
                output_obj["str"] = text_str
            json.dump(output_obj, sys.stdout, indent=4)
        else:
            with redirect_stdout(sys.stderr):
                output.primary(f"{label} (dec): {text}")
                output.primary(f"{label} (hex): {text_hex}")
                output.primary(f"{label} (raw): {text_raw}")
                output.primary(f"{label} (b64): {text_b64}")
                output.primary(f"{label} (url): {text_b64_url}")
                if text_str is not None:
                    output.primary(f"{label} (str): {text_str}")
    else:
        with open(filename, "wb") as file:
            file.write(text_raw)


# From a tip I saw here: https://stackoverflow.com/questions/3431825/generating-an-md5-checksum-of-a-file
def file_checksum(filename):
    hash_sha256 = hashlib.sha256()
    with open(filename, "rb") as f:
        for chunk in iter(partial(f.read, 4096), b""):
            hash_sha256.update(chunk)
    return hash_sha256.hexdigest()


def module_root(m):
    mpath = Path(m.__file__)
    if mpath.stem == "__init__":
        mpath = mpath.parent
    return mpath


def copy_resource(package, res, dest):
    with resources.open_binary(package, res) as src, \
            open(Path(dest)/res, "wb") as dst:
        copyfileobj(src, dst)


def copy_resource_tree(package, dest):
    package_name = package.__name__.split(".")[-1]
    dest_subdir = Path(dest)/package_name
    dest_subdir.mkdir(mode=0o755, exist_ok=True)
    for x in resources.contents(package):
        if resources.is_resource(package, x):
            copy_resource(package, x, dest_subdir)
        elif x != "__pycache__":
            subpackage = import_module(f".{x}", package.__name__)
            copy_resource_tree(subpackage, dest_subdir)
